#!@l_prefix@/bin/perl
##
##  registry-db.pl - OpenPKG Registry - DataBase utility
##  Copyright (c) 2000-2005 OpenPKG Foundation e.V. <http://openpkg.net/>
##  Copyright (c) 2000-2005 Ralf S. Engelschall <http://engelschall.com/>
##
##  Permission to use, copy, modify, and distribute this software for
##  any purpose with or without fee is hereby granted, provided that
##  the above copyright notice and this permission notice appear in all
##  copies.
##
##  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
##  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
##  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
##  IN NO EVENT SHALL THE AUTHORS AND COPYRIGHT HOLDERS AND THEIR
##  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
##  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
##  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
##  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
##  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
##  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
##  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
##  SUCH DAMAGE.
##

require 5.008;
use strict;

use Data::Dumper;
use OSSP::cfg;
use DBI;
use DBD::Pg;
use Getopt::Long;

#   program name, version and date
my $progname="registry-db.pl";
my $progvers="0.7.2";
my $progdate="09-Aug-2006";

#   configure data dumper
$Data::Dumper::Purity = 1;
$Data::Dumper::Indent = 1;
$Data::Dumper::Terse  = 1;

#   determine path to OpenPKG instance
my $PREFIX='@l_prefix@';
$PREFIX=$ENV{OPENPKG_PREFIX} if ($ENV{OPENPKG_PREFIX} ne "");

#   registry common code
my $pm = "registry.pm";
$pm = "$PREFIX/libexec/openpkg-registry/registry.pm" unless($PREFIX =~ m/^@.*@$/);
require $pm; import registry;

#   derive data version from program version
my $datavers={};
$datavers->{value} = $progvers;
if ($datavers->{value} =~ m/^[^0-9]*([0-9]+)\.([0-9]+)\.([0-9]+)/) {
    $datavers->{major}=$1;
    $datavers->{minor}=$2;
    $datavers->{patch}=$3;
}
else {
    die "cannot derive datavers major.minor.patch triple from progvers=" . $progvers;
}

#   options and defaults
my $opt = {};
$opt->{query}   = "";
$opt->{qlist}   = 0;
$opt->{prefix}  = $PREFIX;
$opt->{config}  = &registry::getcfgfilename($PREFIX);
$opt->{debug}   = 0;
$opt->{help}    = 0;
$opt->{version} = 0;

sub debug () {
    my ($txt) = @_;
    printf STDERR "$progname:DEBUG: %s\n", $txt if($opt->{debug});
}

sub warning () {
    my ($txt) = @_;
    printf STDERR "$progname:WARNING: %s\n", $txt;
}

sub error () {
    my ($txt) = @_;
    printf STDERR "$progname:ERROR: %s\n", $txt;
    exit(1);
}

sub version ()
{
    printf "$progname:INFO: version=%s date=%s\n", $progvers, $progdate;
    exit(0);
}

sub usage ()
{
    my ($level) = @_;
    my $txt = qq{
        $progname:$level: Usage: $progname [options] command
        Available options:
         -q|query=s  query specific configuration directive
         -l|qlist    list available query directives
         -p|prefix=s specifiy prefix for data ($opt->{prefix})
         -C|config=s specifiy config file ($opt->{config})
         -D|debug    turn on debugging
         -H|help     print out this usage page
         -V|version  print version
        Available commands:
         setup       setup/upgrade database
         erase       erase all data from database
         backup      backup database
         restore     restore database
         defcfg      show default configuration
         usecfg      show used configuration
         effcfg      show effective configuration
    };
    $txt =~ s/^\n//s; $txt =~ s/^[ ]{8}//mg; $txt =~ s/ *$//mg;
    if ($level eq "INFO") {
        print $txt;
        exit(0);
    }
    print STDERR $txt;
    exit(1);
}

Getopt::Long::Configure("bundling");
my $result = GetOptions(
    'q|query=s'      => \$opt->{query},
    'l|qlist'        => \$opt->{qlist},
    'p|prefix=s'     => \$opt->{prefix},
    'C|config=s'     => \$opt->{config},
    'D|debug'        => \$opt->{debug},
    'H|help'         => \$opt->{help},
    'V|version'      => \$opt->{version}
) || &usage("ERROR");
&version() if ($opt->{version});
&usage("INFO") if ($opt->{help});
( $#ARGV == 0 ) || &usage("ERROR");

#   check data prefix
if (   $ARGV[0] eq "setup" 
    or $ARGV[0] eq "erase"
    or $ARGV[0] eq "backup"
    or $ARGV[0] eq "restore") {
    &error("required data prefix empty") if ($opt->{prefix} eq "");
    &error("required data prefix unset") if ($opt->{prefix} =~ m/^@.*@$/);
}

#   handle configuration
my $defcfgtxt = &registry::getdefcfgtxt($opt->{prefix});
my $defcfg    = &registry::parsecfgtxt($defcfgtxt);
my $usecfgtxt = &registry::readcfgtxtfile($opt->{config});
my $usecfg    = &registry::parsecfgtxt($usecfgtxt);
my $cfg       = &registry::mergecfg($defcfg, $usecfg);

#   database handle and scratch variables
#
my $sql; # scratch SQL string database operations
my $rv;  # scratch return value for database operations
my $rc;  # scratch return code for database operations
my $sth; # scratch statement handle for database operations
my $msg; # scratch variable for message fiddling
my $dbh; # database handle
my $dbv; # database version

if    ($ARGV[0] eq "setup") {
    &setup();
}
elsif ($ARGV[0] eq "erase") {
    &erase();
}
elsif ($ARGV[0] eq "backup") {
    &backup();
}
elsif ($ARGV[0] eq "restore") {
    &restore();
}
elsif ($ARGV[0] eq "defcfg") {
    if ($opt->{qlist}) {
        &querylist(0, $defcfg);
    }
    if ($opt->{query}) {
        &querydirective($defcfg, $opt->{query});
    }
    if (not $opt->{qlist} and not $opt->{query}) {
        if ($opt->{debug}) {
            print Dumper($defcfg);
        }
        else {
            print $defcfgtxt;
        }
    }
}
elsif ($ARGV[0] eq "usecfg") {
    if ($opt->{qlist}) {
        &querylist(0, $usecfg);
    }
    if ($opt->{query}) {
        &querydirective($usecfg, $opt->{query});
    }
    if (not $opt->{qlist} and not $opt->{query}) {
        if ($opt->{debug}) {
            print Dumper($usecfg);
        }
        else {
            print $usecfgtxt;
        }
    }
}
elsif ($ARGV[0] eq "effcfg") {
    if ($opt->{qlist}) {
        &querylist(0, $cfg);
    }
    if ($opt->{query}) {
        &querydirective($cfg, $opt->{query});
    }
    if (not $opt->{qlist} and not $opt->{query}) {
        if ($opt->{debug}) {
            print Dumper($cfg);
        }
        else {
            &querylist(1, $cfg);
        }
    }
}
exit(0);

sub querylist ($$)
{
    my ($vflag, $c, $pre) = @_;
    foreach my $k (sort keys %$c) {
        if (ref $c->{$k} eq "HASH" or ref $c->{$k} eq "ARRAY") {
            &querylist($vflag, $c->{$k}, $pre . "/" . $k);
        }
        else {
            if ($vflag) {
                my $v = $c->{$k};
                $v =~ s|\\|\\\\|sg;
                $v =~ s|\n|\\n|sg;
                $v =~ s|\"|\\"|sg;
                printf "%s=\"%s\"\n", $pre . "/" . $k, $v;
            }
            else {
                printf "%s\n", $pre . "/" . $k;
            }
        }
    }
}

sub querydirective ($$)
{
    my ($c, $tok) = @_;
    my $p = join "", map { "->{$_}" if($_) } split "/", $tok;
    print eval "\$c$p";
}

sub backup ()
{
    my $host       = $cfg->{db}->{registry}->{host};
    my $username   = $cfg->{db}->{registry}->{username};
    my $password   = $cfg->{db}->{registry}->{password};
    my $database   = $cfg->{db}->{registry}->{database};
    my $dumpfile   = $cfg->{db}->{registry}->{dumpfile};
    my $pg_dump    = $cfg->{db}->{registry}->{pg_dump};
    &error("required password empty") if ($password eq "");
    $ENV{PGPASSWORD} = $password;
    system "$pg_dump </dev/null --host=$host --username=$username $database >$dumpfile";
    if ($? == -1) {
        &error("pg_dump executing failed: $!");
    }
    elsif ($? & 127) {
        &error("pg_dump died with signal %d", ($? & 127));
    }
    elsif($? >> 8) {
        &error("pg_dump failed with error code %d", ($? >> 8));
    }
}

sub restore ()
{
    my $host       = $cfg->{db}->{registry}->{host};
    my $username   = $cfg->{db}->{registry}->{username};
    my $password   = $cfg->{db}->{registry}->{password};
    my $database   = $cfg->{db}->{registry}->{database};
    my $dumpfile   = $cfg->{db}->{registry}->{dumpfile};
    my $psql       = $cfg->{db}->{registry}->{psql};
    &error("required password empty") if ($password eq "");
    $ENV{PGPASSWORD} = $password;
    my $pipe = new IO::File "| $psql -h $host -U $username $database";

    #   restore requires tables in certain order to obey constraints
    #   it is assumed the dumpfile does not fit in memory, so multiple
    #   read passes are required - goodbye, STDIN.
    #   reg_config is intentionally skipped
    #
    for my $table (qw{
        reg_user reg_token reg_instance reg_statistic
        pkg_ids
        nfv_flags nfv_names nfv_versions nfv
        nvr_names nvr_releases nvr_versions nvr
    }) {
        &debug("restore searching dump for table=$table");
        my $io = new IO::File "<$dumpfile";
        &error("cannot open \"$dumpfile\" for reading ($!)") if(not defined $io);
        my $copymode = 0;
        my $rows = 0;
        while (<$io>) {
            $copymode = 1 if /^COPY $table /;
            if ($copymode) {
                print $pipe $_;
                $rows++;
            }
            $copymode = 0 if /^\\\./;
        };
        $io->close();
        &debug("restore found table=$table, rows=$rows");
        &warning("restore found no data for table=$table") if ($rows == 0);
    }
    $pipe->close();
}

sub businesscategories ($)
{
    my $bcfile = $cfg->{db}->{registry}->{businesscategories};
    &debug("loading business categories from file \"$bcfile\"");
    my $io = new IO::File "<$bcfile";
    my $data;
    while (<$io>) {
        chomp;
        $data = $dbh->quote($_);
        $sql = sprintf("INSERT INTO reg_businesscategories SELECT %s WHERE NOT EXISTS ( SELECT bc FROM reg_businesscategories WHERE (bc = %s ) );", $data, $data);
        &dbdo($dbh, $sql) || die;
    };
    $io->close();
}

sub countrycodes ($)
{
    my $ccfile = $cfg->{db}->{registry}->{countrycodes};
    &debug("loading country codes from file \"$ccfile\"");
    my $io = new IO::File "<$ccfile";
    my $data;
    my $copymode = 0;
    while (<$io>) {
        chomp;
        $data = $_;
        my ($human, $cc) = split /;/, $data;
        $cc = $dbh->quote($cc);
        $human = $dbh->quote($human);
        $sql = sprintf("INSERT INTO reg_countrycodes SELECT %s, %s WHERE NOT EXISTS ( SELECT cc FROM reg_countrycodes WHERE (cc = %s ) );", $cc, $human, $cc);
        &dbdo($dbh, $sql) || die;
        $sql = sprintf("UPDATE reg_countrycodes SET human = %s WHERE ( cc = %s );", $human, $cc);
        &dbdo($dbh, $sql) || die;
    };
    $io->close();
}

sub dbopen($)
{
    my ($db) = @_;
    my $dbi;
    $dbi = undef;

    if    ($db eq "template1") {
        my $host     = $cfg->{db}->{registry}->{host};
        my $username = $cfg->{db}->{registry}->{superuser}->{username};
        my $password = $cfg->{db}->{registry}->{superuser}->{password};
        &debug("processing Database \"$db\" host=\"$host\", user=\"$username\", pass=\"$password\"");
        $dbi = DBI->connect (
            "DBI:Pg:dbname=template1".";host=$host", $username, $password,
            {
                PrintError => 0,
                PrintWarn  => 0,
                RaiseError => 0,
                AutoCommit => 1
            }
        );
    }
    elsif ($db eq "registry/superuser") {
        my $tablespace = $cfg->{db}->{registry}->{tablespace};
        my $host       = $cfg->{db}->{registry}->{host};
        my $username   = $cfg->{db}->{registry}->{superuser}->{username};
        my $password   = $cfg->{db}->{registry}->{superuser}->{password};
        &debug("processing Database \"$db\" tablespace=\"$tablespace\" host=\"$host\", user=\"$username\", pass=\"$password\"");
        $dbi = DBI->connect (
            "DBI:Pg:dbname=$tablespace;host=$host", $username, $password,
            {
                PrintError => 0,
                PrintWarn  => 0,
                RaiseError => 0,
                AutoCommit => 1
            }
        );
    }
    elsif ($db eq "registry") {
        my $tablespace = $cfg->{db}->{registry}->{tablespace};
        my $host       = $cfg->{db}->{registry}->{host};
        my $username   = $cfg->{db}->{registry}->{username};
        my $password   = $cfg->{db}->{registry}->{password};
        &debug("processing Database \"$db\" tablespace=\"$tablespace\" host=\"$host\", user=\"$username\", pass=\"$password\"");
        $dbi = DBI->connect (
            "DBI:Pg:dbname=$tablespace;host=$host", $username, $password,
            {
                PrintError => 0,
                PrintWarn  => 0,
                RaiseError => 0,
                AutoCommit => 1
            }
        );
    }
    elsif ($db eq "session") {
        my $dbfile = $cfg->{db}->{session}->{dbfile};
        &debug("processing Database \"$db\" dbfile=\"$dbfile\"");
        $dbi = DBI->connect (
            "DBI:SQLite:dbname=$dbfile", "", "",
            {
                PrintError => 0,
                PrintWarn  => 0,
                RaiseError => 0,
                AutoCommit => 1
            }
        );
    }
    else {
        die "dbopen does not know how to handle db=\"".$db."\"";
    }
    return $dbi;
}

sub dbdo ($$)
{
    my ($dbh, $sql, $ignore) = @_;
    my ($rv, $txt, $msg, $ok, $expectederror);

    $sql =~ s/ *\n+$//s;
    $rv = $dbh->do($sql);
    $ok = 0; $ok = 1 if ($rv eq '0E0' or $rv >= 1);
    $msg = $dbh->errstr;
    $msg =~ s/ *\n+$//s;
    $expectederror = 0; $expectederror = 1 if ($msg =~ m/$ignore/);
    $txt = sprintf "\n    sql=\"%s\"\n    ret=\"%s\"\n    msg=\"%s\"\n", $sql, $rv, $msg;
    if (not $ok and not $expectederror) {
        printf STDERR "$progname:ERROR: %s\n", $txt;
        return 0;
    }
    &debug($txt);
    return 1;
}

sub erase ()
{
    $dbh = &dbopen("template1");
    die "ERROR:$0: DataBase error: ".$DBI::errstr."\n" if (not defined $dbh);

    $sql = sprintf "DROP DATABASE %s;",
        $cfg->{db}->{registry}->{database};
    &dbdo($dbh, $sql,
        'database .* does not exist') || die;

    $sql = sprintf "DROP TABLESPACE %s;",
        $cfg->{db}->{registry}->{tablespace};
    &dbdo($dbh, $sql,
        'tablespace .* does not exist') || die;

    $sql = sprintf "DROP USER %s;",
        $cfg->{db}->{registry}->{username};
    &dbdo($dbh, $sql,
        '(user|role) .* does not exist') || die;

    my $dbfile = $cfg->{db}->{session}->{dbfile};
    &debug("truncate session database dbfile=\"$dbfile\"");
    truncate $dbfile, 0;
}

sub setup ()
{
    #   create user, tablespace and database
    #
    $dbh = &dbopen("template1");
    die "ERROR:$0: DataBase error: ".$DBI::errstr."\n" if (not defined $dbh);

    $sql = sprintf "CREATE USER %s ENCRYPTED PASSWORD %s NOCREATEDB NOCREATEUSER;",
        $cfg->{db}->{registry}->{username},
        $dbh->quote($cfg->{db}->{registry}->{password});
    &dbdo($dbh, $sql,
        '(user|role) .* already exists') || die;

    $sql = sprintf "CREATE TABLESPACE %s OWNER %s LOCATION %s;",
        $cfg->{db}->{registry}->{tablespace},
        $cfg->{db}->{registry}->{username},
        $dbh->quote($cfg->{db}->{registry}->{location});
    &dbdo($dbh, $sql,
        'tablespace .* already exists') || die;

    $sql = sprintf "CREATE DATABASE %s OWNER %s TABLESPACE %s;",
        $cfg->{db}->{registry}->{database},
        $cfg->{db}->{registry}->{username},
        $cfg->{db}->{registry}->{tablespace};
    &dbdo($dbh, $sql,
        'database .* already exists') || die;

    $dbh->disconnect(); undef $dbh;

    #   enable Pg/SQL
    #
    $dbh = &dbopen("registry/superuser");
    die "ERROR:$0: DataBase error: ".$DBI::errstr."\n" if (not defined $dbh);

    $sql = q{
        CREATE FUNCTION plpgsql_call_handler () RETURNS LANGUAGE_HANDLER AS '$libdir/plpgsql' LANGUAGE C;
        COMMENT ON FUNCTION plpgsql_call_handler () IS 'PL/pgSQL Call Handler';
        CREATE TRUSTED PROCEDURAL LANGUAGE plpgsql HANDLER plpgsql_call_handler;
    };
    &dbdo($dbh, $sql,
        'function .* already exists') || die;

    $dbh->disconnect(); undef $dbh;

    #   session
    #
    $dbh = &dbopen("session");
    die "ERROR:$0: DataBase error: ".$DBI::errstr."\n" if (not defined $dbh);
    chmod oct("0664"), $cfg->{db}->{session}->{dbfile};
    while (&upgradesession()) {};
    $dbh->disconnect(); undef $dbh;

    #   registry
    #
    $dbh = &dbopen("registry");
    die "ERROR:$0: DataBase error: ".$DBI::errstr."\n" if (not defined $dbh);
    while (&upgraderegistry()) {};
    &businesscategories();
    &countrycodes();
    $dbh->disconnect(); undef $dbh;

}

sub upgradesession()
{
    ($dbv, $rv) = undef;

    $sql="SELECT * FROM config;";
    $rv = $dbh->selectall_hashref($sql, "key");
    if (not defined %{$rv}->{version}) {
        $dbv = { 'version' => { 'key' => 'version', 'value' => '0.0.0' } }
    }
    else {
        $dbv = $rv;
    }

    if ($dbv->{version}->{value} =~ m/^([0-9]+)\.([0-9]+)\.([0-9]+)$/) {
        $dbv->{version}->{major}=$1;
        $dbv->{version}->{minor}=$2;
        $dbv->{version}->{patch}=$3;
    }
    else {
        die "cannot derive database major.minor.patch triple from value=".$dbv->{version}->{value};
    }

    #   program data version and database version already match
    #
    if    (    $dbv->{version}->{major} == $datavers->{major}
           and $dbv->{version}->{minor} == $datavers->{minor}
           and $dbv->{version}->{patch} == $datavers->{patch}) {
       &debug("database=" . $dbv->{version}->{major} . "."
                          . $dbv->{version}->{minor} . "."
                          . $dbv->{version}->{patch}
                          . " is at exactly program datavers="
                          . $datavers->{major} . "."
                          . $datavers->{minor} . "."
                          . $datavers->{patch});
        return 0;
    }

    #   0.0.0 upgrade to 0.0.1 - initialization from scratch
    #
    elsif (    $dbv->{version}->{major} == 0
           and $dbv->{version}->{minor} == 0
           and $dbv->{version}->{patch} == 0) {
        $sql = &schemasession1();
        &dbdo($dbh, $sql) || die;
        $sql = &schemasession2();
        &dbdo($dbh, $sql) || die;
        $sql = "INSERT INTO config VALUES ( 'version', '0.0.1' );";
        &dbdo($dbh, $sql) || die;
        return 1;
    }

    #   0.0.x upgrade to 0.0.24
    #
    elsif (    $dbv->{version}->{major} == 0
           and $dbv->{version}->{minor} == 0
           and $dbv->{version}->{patch} <  24 ) {
        $sql = &schemasession3();
        &dbdo($dbh, $sql) || die;
        $sql = "UPDATE config SET value = '0.0.24' WHERE ( key = 'version' );";
        &dbdo($dbh, $sql) || die;
        return 1;
    }

    #   0.0.24 upgrade to 0.7.0
    #
    elsif (    $dbv->{version}->{major} == 0
           and $dbv->{version}->{minor} <  7  ) {
        $sql = "UPDATE config SET value = '0.7.0' WHERE ( key = 'version' );";
        &dbdo($dbh, $sql) || die;
        return 1;
    }

    #   program data version and database version only differ in patchlevel - just uprev the number
    #
    elsif (    $dbv->{version}->{major} == $datavers->{major}
           and $dbv->{version}->{minor} == $datavers->{minor}
           and $dbv->{version}->{patch} <  $datavers->{patch}) {
        &debug("database=" . $dbv->{version}->{major} . "."
                          . $dbv->{version}->{minor} . "."
                          . $dbv->{version}->{patch}
                          . " is at program major/minor datavers="
                          . $datavers->{major} . "."
                          . $datavers->{minor} . "."
                          . $datavers->{patch});
        $sql = sprintf("UPDATE config SET value = '%s.%s.%s' WHERE ( key = 'version' );",
                            $datavers->{major},
                            $datavers->{minor},
                            $datavers->{patch});
        &dbdo($dbh, $sql) || die;
        return 1;
    }

    #   unsupported upgrade case
    #
    else {
        print "program version=" . $progvers . " cannot handle database version=" . $dbv->{version}->{value} . "\n";
        print "database=" . $dbv->{version}->{major} . "."
                          . $dbv->{version}->{minor} . "."
                          . $dbv->{version}->{patch}
                          . " cannot be upgraded by this program which uses datavers="
                          . $datavers->{major} . "."
                          . $datavers->{minor} . "."
                          . $datavers->{patch}
                          . "\n";
        return 0;
    }
}

sub upgraderegistry()
{
    $dbv = {};

    $sql="SELECT * FROM reg_config;";
    $rv = $dbh->selectall_hashref($sql, "key");
    if (not defined %{$rv}->{version}) {
        $dbv = { 'version' => { 'key' => 'version', 'value' => '0.0.0' } }
    }
    else {
        $dbv = $rv;
    }

    if ($dbv->{version}->{value} =~ m/^([0-9]+)\.([0-9]+)\.([0-9]+)$/) {
        $dbv->{version}->{major}=$1;
        $dbv->{version}->{minor}=$2;
        $dbv->{version}->{patch}=$3;
    }
    else {
        die "cannot derive database major.minor.patch triple from value=".$dbv->{version}->{value};
    }

    #   program data version and database version match
    #
    if    (    $dbv->{version}->{major} == $datavers->{major}
           and $dbv->{version}->{minor} == $datavers->{minor}
           and $dbv->{version}->{patch} == $datavers->{patch}) {
        &debug("database=" . $dbv->{version}->{major} . "."
                          . $dbv->{version}->{minor} . "."
                          . $dbv->{version}->{patch}
                          . " is at exactly program datavers="
                          . $datavers->{major} . "."
                          . $datavers->{minor} . "."
                          . $datavers->{patch});
        return 0;
    }

    #   0.0.0 upgrade to 0.1.0 - initialization from scratch
    #
    elsif (    $dbv->{version}->{major} == 0
           and $dbv->{version}->{minor} == 0
           and $dbv->{version}->{patch} == 0) {
        $sql = "";
        $sql .= &schemaregistry();
        $sql .= &funcregistry();
        $sql .= "INSERT INTO reg_config VALUES ( 'version', '0.1.0' );";
        &dbdo($dbh, $sql) || die;
        return 1;
    }


    #   0.1.x upgrade to 0.2.0
    #
    elsif (    $dbv->{version}->{major} == 0
           and $dbv->{version}->{minor} <  2  ) {
        $sql = "UPDATE reg_config SET value = '0.2.0', mtime = now() WHERE ( key = 'version' );";
        &dbdo($dbh, $sql) || die;
        return 1;
    }

    #   0.2.x upgrade to 0.3.0
    #
    elsif (    $dbv->{version}->{major} == 0
           and $dbv->{version}->{minor} == 2  ) {
        $sql = "";
        $sql .= q{
            ALTER TABLE reg_statistic ALTER COLUMN time_first SET DEFAULT CURRENT_TIMESTAMP;
            ALTER TABLE reg_statistic ALTER COLUMN time_last SET DEFAULT CURRENT_TIMESTAMP;
            ALTER TABLE reg_statistic ALTER COLUMN count SET DEFAULT 1;
        };
        $sql .= &funcregistry();
        $sql .= "UPDATE reg_config SET value = '0.3.0', mtime = now() WHERE ( key = 'version' );";
        &dbdo($dbh, $sql) || die;
        return 1;
    }

    #   0.3.x upgrade to 0.4.0
    #
    elsif (    $dbv->{version}->{major} == 0
           and $dbv->{version}->{minor} == 3  ) {
        $sql = "";
        $sql .= q{
            ALTER TABLE reg_user ADD COLUMN fullname TEXT;
        };
        $sql .= "UPDATE reg_config SET value = '0.4.0', mtime = now() WHERE ( key = 'version' );";
        &dbdo($dbh, $sql) || die;
        return 1;
    }

    #   0.4.x upgrade to 0.5.0
    #
    elsif (    $dbv->{version}->{major} == 0
           and $dbv->{version}->{minor} == 4) {
        $sql = "";
        $sql .= q{
            ALTER TABLE reg_instance DROP CONSTRAINT reg_instance_registry_orel_check;
            ALTER TABLE reg_instance ADD CONSTRAINT reg_instance_registry_orel_check
                CHECK (registry_orel ~ '^[^-]+-((CURRENT|CURRENT-[0-9]+|[^.-]+-STABLE|[^.-]+-STABLE-[0-9]+|[^.-]+\.[^.-]+-SOLID|[^.-]+\.[^.-]+-SOLID-[0-9]+|[^.-]+\.[^.-]+-RELEASE).*|[0-9]+\.[0-9]+)$');
        };
        $sql .= "UPDATE reg_config SET value = '0.5.0', mtime = now() WHERE ( key = 'version' );";
        &dbdo($dbh, $sql) || die;
        return 1;
    }

    #   0.5.x upgrade to 0.6.0
    #
    elsif (    $dbv->{version}->{major} == 0
           and $dbv->{version}->{minor} == 5) {
        $sql = "";
        $sql .= &schemaregistry2();
        $sql .= "UPDATE reg_config SET value = '0.6.0', mtime = now() WHERE ( key = 'version' );";
        &dbdo($dbh, $sql) || die;
        return 1;
    }

    #   0.6.x upgrade to 0.7.0
    #
    elsif (    $dbv->{version}->{major} == 0
           and $dbv->{version}->{minor} == 6  ) {
        $sql = "";
        $sql .= &schemaregistry3();
        $sql .= q{
            ALTER TABLE reg_user ADD COLUMN organization TEXT;
            ALTER TABLE reg_user ADD COLUMN businesscategory TEXT REFERENCES reg_businesscategories (bc);
            ALTER TABLE reg_user ADD COLUMN addresslabel TEXT;
            ALTER TABLE reg_user ADD COLUMN country TEXT REFERENCES reg_countrycodes (cc);
            ALTER TABLE reg_user ADD COLUMN timezone TEXT;
            ALTER TABLE reg_user ADD COLUMN telephonenumber TEXT;
            ALTER TABLE reg_user ADD COLUMN mobilenumber TEXT;
            ALTER TABLE reg_instance ADD COLUMN
                registry_link TEXT
                              REFERENCES reg_token (id)
                              -- reference to linked token
                              -- [0ee9713e-2560-11db-a5f5-001279cd6fab]
                              ;
        };
        $sql .= "UPDATE reg_config SET value = '0.7.0', mtime = now() WHERE ( key = 'version' );";
        &dbdo($dbh, $sql) || die;
        return 1;
    }

    #   program data version and database version only differ in patchlevel - just uprev the number
    #
    elsif (    $dbv->{version}->{major} == $datavers->{major}
           and $dbv->{version}->{minor} == $datavers->{minor}
           and $dbv->{version}->{patch} <  $datavers->{patch}) {
        &debug("database=" . $dbv->{version}->{major} . "."
                          . $dbv->{version}->{minor} . "."
                          . $dbv->{version}->{patch}
                          . " is at program major/minor datavers="
                          . $datavers->{major} . "."
                          . $datavers->{minor} . "."
                          . $datavers->{patch});
        $sql = sprintf("UPDATE reg_config SET value = '%s.%s.%s', mtime = now() WHERE ( key = 'version' );",
                            $datavers->{major},
                            $datavers->{minor},
                            $datavers->{patch});
        &dbdo($dbh, $sql) || die;
        return 1;
    }

    #   unsupported upgrade case
    #
    else {
        print "program version=" . $progvers . " cannot handle database version=" . $dbv->{version}->{value} . "\n";
        print "database=" . $dbv->{version}->{major} . "."
                          . $dbv->{version}->{minor} . "."
                          . $dbv->{version}->{patch}
                          . " cannot be upgraded by this program which uses datavers="
                          . $datavers->{major} . "."
                          . $datavers->{minor} . "."
                          . $datavers->{patch}
                          . "\n";
        return 0;
    }
}

sub schemasession1()
{
    return <<'EOT'
    -- OpenPKG configuration
    CREATE TABLE config (
        key           TEXT 
                      NOT NULL
                      PRIMARY KEY,
                      -- generic key
                      -- [version]
        value         TEXT
                      NOT NULL
                      -- generic value
                      -- [0.8.15]
    );
EOT
}

sub schemasession2()
{
    return <<'EOT'
    -- OpenPKG session
   CREATE TABLE session (
        id             CHAR(36) NOT NULL PRIMARY KEY,
                       -- Session Identifier (OSSP::uuid based)
                       -- [75468374-8721-11d9-bee4-0002a5489f14]
        a_session      BLOB NOT NULL
                       -- Session Data (Storage based)
                       -- [XXXXXXX...]
    );
EOT
}

sub schemasession3()
{
    return <<'EOT'
    -- OpenPKG cache
   CREATE TABLE cache (
        url            TEXT NOT NULL
                       PRIMARY KEY
                       UNIQUE ON CONFLICT REPLACE,
                       -- URL to be cached
                       -- [http://meta.openpkg.org/]
        content_type   TEXT,
                       -- HTTP content type
                       -- [text/html]
        expires        INTEGER NOT NULL,
                       -- Invalidation time in sec since 1970
                       -- [121343]
        content        BLOB
                       -- Session Data (Storage based)
                       -- [...]
    );
EOT
}

sub schemaregistry()
{
    return <<'EOT'
    -- OpenPKG configuration
    CREATE TABLE reg_config (
        key           TEXT
                      UNIQUE NOT NULL,
                      -- generic key
                      -- [version]
        value         TEXT
                      NOT NULL,
                      -- generic value
                      -- [0.8.15]
        mtime         TIMESTAMP (0) WITHOUT TIME ZONE
                      NOT NULL
                      DEFAULT CURRENT_TIMESTAMP,
                      -- time of last modification
                      -- [2005-11-08 00:00:00]
        PRIMARY KEY (key)
    );

    -- OpenPKG user
    CREATE TABLE reg_user (
        username      TEXT
                      NOT NULL
                      CHECK(username ~ '^.+@.+$'),
                      -- OSSP ase user
                      -- [one@example.com]
        heartbeat     TIMESTAMP (0) WITHOUT TIME ZONE
                      NOT NULL
                      DEFAULT CURRENT_TIMESTAMP,
                      -- time of last registry.cgi usage
                      -- [2005-11-08 00:00:00]
        PRIMARY KEY (username)
    );

    -- OpenPKG instance
    CREATE TABLE reg_instance (
        uuid_registry TEXT
                      UNIQUE NOT NULL
                      CHECK (uuid_registry ~ '^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$'),
                      -- <prefix>/etc/openpkg/uuid:UUID_REGISTRY
                      -- [d98267da-0874-11d9-b8f6-0002a5489f14]
        uuid_instance TEXT
                      NOT NULL
                      CHECK (uuid_instance ~ '^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$'),
                      -- <prefix>/etc/openpkg/uuid:UUID_INSTANCE
                      -- [1169823e-22db-3547-9f72-20c2ef18619f]
        uuid_platform TEXT
                      NOT NULL
                      CHECK (uuid_platform ~ '^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$'),
                      -- <prefix>/etc/openpkg/uuid:UUID_PLATFORM
                      -- [4a27ce02-730e-3cbf-a3fd-c32042c5f619]
        registry_desc TEXT
                      NOT NULL,
                      -- human-friendly description
                      -- [openpkg://host.domain.invalid/openpkg]
        registry_user TEXT
                      NOT NULL
                      REFERENCES reg_user (username),
                      -- reference to associated user
                      -- [one@example.com]
        registry_plat TEXT
                      CHECK (registry_plat ~ '^[^-]+-[^-]+$'),
                      -- output from openpkg rpm --eval '%{l_platform -p}'
                      -- [ix86-freebsd5.4]
        registry_orel TEXT
                      CHECK (registry_orel ~ '^[^-]+-((CURRENT|CURRENT-[0-9]+|[^.-]+-STABLE|[^.-]+-STABLE-[0-9]+|[^.-]+\.[^.-]+-SOLID|[^.-]+\.[^.-]+-SOLID-[0-9]+|[^.-]+\.[^.-]+-RELEASE).*|[0-9]+\.[0-9]+)$'),
                      -- output from openpkg rpm --eval '%{l_openpkg_release}'
                      -- [OpenPKG-2.5]
        registry_date TIMESTAMP (0) WITHOUT TIME ZONE
                      NOT NULL
                      DEFAULT CURRENT_TIMESTAMP,
                      -- time of registration or last reregistration
                      -- [2005-11-08 00:00:00]
        associated    BOOLEAN
                      NOT NULL
                      DEFAULT 'false',
                      -- whether instance is associated to user
                      -- [true]
        PRIMARY KEY (uuid_registry)
    );

    -- Access Statistic
    CREATE TABLE reg_statistic (
        login         TEXT
                      NOT NULL
                      CHECK (   login ~ '^.+@.+'
                             OR login ~ '^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$'),
                      -- reference to either reg_user.username or instance.uuid_registry
                      -- [one@example.com] [d98267da-0874-11d9-b8f6-0002a5489f14]
        addr          TEXT
                      NOT NULL,
                      -- remote peer IP address
                      -- [1.2.3.4]
        time_first    TIMESTAMP (0) WITHOUT TIME ZONE
                      NOT NULL
                      DEFAULT CURRENT_TIMESTAMP,
                      -- time login+addr was seen first
                      -- [2005-11-08 00:00:00]
        time_last     TIMESTAMP (0) WITHOUT TIME ZONE
                      NOT NULL
                      DEFAULT CURRENT_TIMESTAMP,
                      -- time login+addr was seen last
                      -- [2005-11-08 00:00:00]
        count         INTEGER
                      NOT NULL
                      DEFAULT '1',
                      -- number login+addr was seen at all
                      -- [0]
        PRIMARY KEY (login, addr)
    );
EOT
}

sub schemaregistry2()
{
    return <<'EOT'
    -- package identifiers
    CREATE TABLE pkg_ids (
        id            TEXT
                      CHECK (id ~ '^[0-9a-f]{32}$'),
                      -- [d8f831be5def1d1dfb452c0a711e5903]
        type          TEXT
                      NOT NULL,
                      CHECK (type = 'db' OR type = 'rpm' OR type = 'srpm'),
                      -- [db]
                      PRIMARY KEY (id)
    );

    -- package name-version-release
    CREATE TABLE nvr_names (
        name          TEXT,
                      -- [ssmtp]
                      PRIMARY KEY (name)
    );
    CREATE TABLE nvr_versions (
        version       TEXT
                      CHECK (version ~ '^[^-]+$'),
                      -- [2.61]
                      PRIMARY KEY (version)
    );
    CREATE TABLE nvr_releases (
        release       TEXT
                      CHECK (release ~ '^[^-]+$'),
                      -- [20050608]
                      PRIMARY KEY (release)
    );
    CREATE TABLE nvr (
        name          TEXT REFERENCES nvr_names(name) ON DELETE CASCADE,
        version       TEXT REFERENCES nvr_versions(version) ON DELETE CASCADE,
        release       TEXT REFERENCES nvr_releases(release) ON DELETE CASCADE,
                      UNIQUE (name, version, release)
    );
    -- provide name-flag-version
    CREATE TABLE nfv_names (
        name          TEXT,
                      -- [ssmtp]
                      PRIMARY KEY (name)
    );
    CREATE TABLE nfv_flags (
        flag          TEXT,
                      -- [=]
                      PRIMARY KEY (flag)
    );
    CREATE TABLE nfv_versions (
        version       TEXT,
                      -- [2.61-20050608]
                      PRIMARY KEY (version)
    );
    CREATE TABLE nfv (
        name          TEXT REFERENCES nfv_names(name) ON DELETE CASCADE,
        flag          TEXT REFERENCES nfv_flags(flag) ON DELETE CASCADE,
        version       TEXT REFERENCES nfv_versions(version) ON DELETE CASCADE,
                      UNIQUE (name, flag, version)
    );
    -- instance package list
    CREATE TABLE inst_pkg_list (
        ipl_inst        TEXT REFERENCES reg_instance(uuid_registry) ON DELETE CASCADE,
        ipl_pkg_id      TEXT REFERENCES pkg_ids(id) ON DELETE CASCADE,
        ipl_nvr_name    TEXT,
        ipl_nvr_version TEXT,
        ipl_nvr_release TEXT,
                        FOREIGN KEY (ipl_nvr_name, ipl_nvr_version, ipl_nvr_release)
                        REFERENCES nvr(name, version, release) ON DELETE CASCADE,
        ipl_nfv_name    TEXT,
        ipl_nfv_flag    TEXT,
        ipl_nfv_version TEXT,
                        FOREIGN KEY (ipl_nfv_name, ipl_nfv_flag, ipl_nfv_version) 
                        REFERENCES nfv(name, flag, version) ON DELETE CASCADE,
                        UNIQUE (ipl_inst, ipl_pkg_id, ipl_nvr_name, ipl_nvr_version, ipl_nvr_release, ipl_nfv_name, ipl_nfv_flag, ipl_nfv_version)
    );
EOT
}

sub schemaregistry3()
{
    return <<'EOT'
    -- OpenPKG token
    CREATE TABLE reg_token (
        id            TEXT
                      UNIQUE NOT NULL
                      CHECK (id ~ '^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$'),
                      -- UUID v1 token
                      -- [d98267da-0874-11d9-b8f6-0002a5489f14]
        heartbeat     TIMESTAMP (0) WITHOUT TIME ZONE,
                      -- time of last usage
                      -- [2005-11-08 00:00:00]
        username      TEXT
                      NOT NULL
                      REFERENCES reg_user (username) ON DELETE CASCADE,
                      -- reference to associated user
                      -- [one@example.com]
        description   TEXT
                      NOT NULL,
                      -- human-friendly description
                      -- [my personal tomahawk]
        userenable      BOOLEAN
                      NOT NULL
                      DEFAULT 'false',
                      -- whether token is accepted as user replacement
                      -- [true]
        usercount     INTEGER
                      NOT NULL
                      DEFAULT '0',
                      -- number of successful user replacement usages
                      -- [0]
        linkenable    BOOLEAN
                      NOT NULL
                      DEFAULT 'false',
                      -- whether token is accepted as link
        linkcount     INTEGER
                      NOT NULL
                      DEFAULT '0',
                      -- number of successful link usages
                      -- [0]
                      -- [true]
        assocenable   BOOLEAN
                      NOT NULL
                      DEFAULT 'false',
                      -- whether token is accepted for automatic association
        assoccount    INTEGER
                      NOT NULL
                      DEFAULT '0',
                      -- number of successful associations
                      -- [0]
                      -- [true]
        PRIMARY KEY (id)
    );
    -- OpenPKG business categories
    CREATE TABLE reg_businesscategories (
        bc            TEXT
                      UNIQUE NOT NULL,
                      -- Finite picklist of business categories
                      -- [News and Media]
        PRIMARY KEY (bc)
    );
    -- OpenPKG country codes
    CREATE TABLE reg_countrycodes (
        cc            TEXT
                      UNIQUE NOT NULL,
                      -- Finite picklist of country codes from ISO3166
                      -- [DE]
        human         TEXT
                      NOT NULL,
                      -- human readable
                      -- [GERMANY]
        PRIMARY KEY (cc)
    );
EOT
}

sub funcregistry()
{
    return <<'EOT'
    --  Stored Procedure: update statistics and fallback to insert in case this becomes a new entry
    CREATE OR REPLACE FUNCTION reg_statisticlog(TEXT, TEXT)
    RETURNS BOOL LANGUAGE 'plpgsql' AS '
    BEGIN
        UPDATE reg_statistic SET time_last = DEFAULT, count = count+1 WHERE ( login = $1 AND addr = $2 );
        IF NOT FOUND THEN
            INSERT INTO reg_statistic (login, addr) VALUES ($1, $2);
        END IF;
        RETURN FOUND;
    END;
    ';
EOT
}

__END__
##
##  MANUAL PAGE
##

=pod

=head1 NAME

B<registry-db> - OpenPKG Registry DataBase utility

=head1 SYNOPSIS

B<registry-db> [options] command
        Available options:
         [I<-q|query=<directive>>]
         [I<-l|qlist>]
         [I<-p|prefix=<path>>]
         [I<-C|config=<file>>]
         [I<-D|debug>] [I<-H|help>] [I<-V|version>]
        Available commands:
         setup       setup/upgrade database
         erase       erase all data from database
         backup      backup database
         restore     restore database
         defcfg      show default configuration
         usecfg      show used configuration
         effcfg      show effective configuration

=head1 DESCRIPTION

B<registry-db> is the DataBase utility for OpenPKG Registry Servers.
It is used by administrators to setup, erase, backup and restore the
Registry Server and to review the default configuration, the contents
of the used configuration file and the internal result of both being
merged.

=head1 OPTIONS

The following options are available:

=over 4

=item B<-D>, B<--debug>

Print verbose debug messages during operation.

=item B<-H>, B<--help>

Display brief usage message.

=item B<-V>, B<--version>

Print version and exit.

=item B<-C>, B<--config=<file>>

Specify configuration file to use.
Default is C<@l_prefix@/etc/openpkg-registry/registry.cfg>

=item B<-p>, B<--prefix=<path>>

Specify path to prefix.  Default is <@l_prefix>, overriden by environment
variable OPENPKG_PREFIX and finally overridden by this option.

=item B<-q>, B<--query=<directive>>

Query a specific configuration directive.
Combine with B<defcfg>, B<usecfg> or B<effcfg> commands.

=item B<-l>, B<--qlist>

List available query directives.
Combine with B<defcfg>, B<usecfg> or B<effcfg> commands.

=back

=head1 COMMANDS

The following commands are available:

=over 4

=item B<setup>

Sets up DataBase from scratch, upgrades or updates it. Maintains
tablespace, role, database and tables. After successfully running
setup, all relations are fully functional but there is no data in the
DataBase. This is an ideal precondition to start from scratch or running
a restore.

=item B<erase>

Erases the DataBase, destroying data, tables, database, role and
tablespace. NO QUESTIONS ASKED! Ideal precondition to get rid of a
Registry Server or restart a setup from scratch.

=item B<backup>

This creates a DataBase backup using pg_dump(1). The content of the
backup dumpfile is as complete as possible and contains data, schema
and more information. The format is PostgreSQL SQL statements. Very
much plain text to be used for further processing. However, due to
limitations of pg_dump not understanding relationships of tables and the
application using tough table constraints attempts to restore by simply
piping the file into psql(1) is left as an excercise to the reader.
Another limitation is that the tablespace is memorized using absolute
pathname, making it hard to restore to a different setup. These problems
can be circumvented using B<registry-db restore>.

=item B<restore>

This restores a dumpfile previously created with B<registry-db backup>
or a similar usage of pg_dump(1). The content of the backup is examined
in multiple passes, one for each table in application order. Only data
is extracted, everthing else is ignored. It is assumed the DataBase is
empty but has been previously prepared using B<registry-db setup>.

=item B<defcfg>

Without query and qlist options dumps the internal default configuration in
plain text format.  When debugging is turned on, the file is printed in
internal parsed format.

=item B<usecfg>

Without query and qlist options dumps the used configuration file in plain
text format.  When debugging is turned on, the file is printed in internal
parsed format.

=item B<effcfg>

Without query and qlist options dumps the effective configuration as a list of
directives with one directive per line in "key=value" format where the value is double
quoted and has backslashes, dollars, double quotes escaped with backslash and
newlines transformed into backslash+"n" thus suitable to be printed with
sprintf or alike.  When debugging is turned on, the file is printed in
internal parsed format.  The effective configuration is the result of the
default configuration and the used configuration being merged with the latter
overriding the former.

=back

=head1 FILES

Default cfg file C<@l_prefix@/etc/openpkg-registry/registry.cfg>

Default dumpfile C<@l_prefix@/var/openpkg-registry/db.dump>

=head1 SEE ALSO

pg_dump(1), psql(1)

=head1 AUTHOR

Thomas Lotterer E<lt>thl@openpkg.orgE<gt>

=cut

